/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include <stdio.h>
#include <math.h>

#include "zibutil.h"
#include "access.h"
#include "params.h"
#include "triang2.h"
#include "nodes.h"
#include "problem.h"
#include "solve.h"

/*  ======================================================================

    BPX -- Realisierung entsprechend :

    Folkmar A. Bornemann: An Adaptive Multilevel Approach to Parabolic 
                          Equations in Two Space Dimensions,
		          Technical Report TR 91-7, Konrad-Zuse-Zentrum,
			  Berlin (1991)	

    Bemerkungen:
    ------------

       1.  Die Felder S^+, U^+, l sind in Sparsematrix-Technik realisiert,
           dies macht die Verwendung des "indexPlusArray"'s noetig.

       2.  Der Zugriff auf die Punkte in M_k erfolgt indirekt ueber die
           verfeinerte Kante der Tiefe k-1. Dies stellt die Beziehung
           Mittelpunkt <-> Eckpunkt her.

       3.  "MKMinus" bezeichnet das (mengentheoretische) Komplement von
           M_k^+ ("MKPlus") in M_(k+1). Dies entspricht den wiederholten
           Fallunterscheidungen.

     ====================================================================== */										

/*   *************
1.0  Deklarationen
     ************* */

static real	*sPlus = nil, *uPlus = nil, *bpxData = nil;
static int *lArray = nil, *indexPlusArray = nil;
static char *dataNames[] = {"qBar", "areaWeight"};

static real meanInitArea, qData;
real bpxTau = ZERO;
static real bpxTau1 = ONE, bpxTauQ;
static int offset[MAX_DEPTH], offsetPlus, counter, nMinus, nPlus;
static int areaKnown, depth, maxDepth, BY, BX, psiNum;

static int AddInitArea(TR *t), InitArea(TR *t),
		   ArrayAlloc(), bpxPcxMul(int x, int y),
		   InitBpxMul(), ShutDownBPX();
static void	InitZeroUPlus(), ArrayFree();
static real	V(PT *p,int k), ValV(int index, int k0, int k1);

static int bpx_p_l = -1, bpx_p_counter = -1,
           bpx_p_mark = -1, bpx_t_area = -1;

/*   **********************************
2.0  BPX - Initialisierung fuer KASKADE
     ********************************** */

int DefBPXPreCond()
  {
	if (bpxData == nil)
	  {
		bpxData = (real*)NewParamList((ptr)nil,"bpxData", 2, sizeof(real),
							 dataNames, T_REAL, 0,(char **)nil,
							 (ptr)nil,(int (*)(char*,char*,int))nil,
							 (int (*)(char*))nil);
		bpxData[0] = 0.0;
		bpxData[1] = 0.25;
/*
		bpxData[0] = 2791.0;
		bpxData[1] = HALF;
*/
	  }
	return DefPreCondMethod("bpx",InitBpxMul,bpxPcxMul,ShutDownBPX);
  }

static int ShutDownBPX()
  {
	ArrayFree();
	if (bpx_p_l!=-1) ReturnAccess(nodesState->accPoint,bpx_p_l);
	bpx_p_l = -1;
	if (bpx_p_counter!=-1) ReturnAccess(nodesState->accPoint,bpx_p_counter);
	bpx_p_counter = -1;
	if (bpx_p_mark!=-1) ReturnAccess(nodesState->accPoint,bpx_p_mark);
	bpx_p_mark = -1;
	if (bpx_t_area!=-1) ReturnAccess(nodesState->accTriangle,bpx_t_area);
	bpx_t_area = -1;

    return true;
  }

static int InitFields(PT *p)
  {
	IP(p, bpx_p_l) = 0;
	IP(p, bpx_p_counter) = 0;
	IP(p, bpx_p_mark) = 0;
	return true;
  }

/*   *********************************************
3.0  Sparse Matrix Routinen fuer benoetigte Felder
     ********************************************* */

static int 	
ArrayAlloc()
{
	psiNum = 2*(actTriang->noOfPoints) - (actTriang)->noOfInitPoints;

	sPlus = (real*) ZIBAlloc((long) sizeof(real)*psiNum);
	if (sPlus == nil)
		return false;

	uPlus = (real*) ZIBAlloc((long) sizeof(real)*psiNum);
	if (uPlus == nil)
		return false;

	lArray = (int*) ZIBAlloc((long) sizeof(int)*psiNum);
	if (lArray == nil)
		return false;

	indexPlusArray = (int*) ZIBAlloc((long) sizeof(int)*psiNum);
	if (indexPlusArray == nil)
		return false;

	return true;
}

static void
ArrayFree()
{
	if (sPlus!=nil)
	  { ZIBFree((PTR) sPlus); sPlus = nil; }
	if (uPlus!=nil)
	  { ZIBFree((PTR) uPlus); uPlus = nil; }
	if (lArray!=nil)
	  { ZIBFree((PTR) lArray); lArray = nil; }
	if (indexPlusArray!=nil)
	  { ZIBFree((PTR) indexPlusArray); indexPlusArray = nil; }
	return;
}

/*   ***************
4.0  Initialisierung
     *************** */

static int
InitArrays()
{
	if(!areaKnown) {
		ArrayFree();
		if (!ArrayAlloc()) {
			sprintf(globBuf,"Not enough memory (ArrayAlloc)\n");
			ZIBStdOut(globBuf);
			return false;
		}
		meanInitArea = ZERO;
		ApplyT(AddInitArea, initial);
		meanInitArea /= (actTriang->noOfInitTriangles);
		qData = (1.0 + bpxTau * bpxData[0]) * meanInitArea;

		ApplyT(InitArea, all);
	}
	InitZeroUPlus();

	return true;
}

/*   ------------------------------
4.1  Hilfsroutinen: Initialisierung
     ------------------------------ */

static void
InitZeroUPlus()
{
	int		i;

	for (i = 0; i < psiNum; i++) 
		uPlus[i] = ZERO;
	return;
}

static int
InitPoint(PT *p)
{
	uPlus[p->indexP] = RA(p, BX);
	IP(p,bpx_p_l) = maxDepth;
	IP(p,bpx_p_counter) = p->indexP;

	return true;
}

static int
AddInitArea(TR *t)
{
	meanInitArea += RT(t,bpx_t_area);

	return true;
}

static int 
InitArea(TR *t)
{
	PT	      **pts = &(t->p1);
	TR	       *tInit;
	real            area;
	int		i;

	tInit = t;
        while ((tInit->father) != nil)
                tInit = tInit->father;

	area = RT(t,bpx_t_area) * bpxData[1] / RA(tInit,bpx_t_area);

	for (i = 0; i< 3; i++)
		sPlus[pts[i]->indexP] += area;

	return true;
}

static int UnmarkPoints(TR *t)
  {
	PT **pts = &(t->p1);
	int i;

	for (i = 0; i < 3; i++) IP(pts[i],bpx_p_mark) = false;
	return true;
  }

static int ComputeArea(TR *t)
  {
	real area;
	PT *p1=t->p1, *p2=t->p2, *p3=t->p3;

	if (RA(t, bpx_t_area)>0) return true;
	area  = (p1->x) * ((p2->y) - (p3->y));
	area += (p2->x) * ((p3->y) - (p1->y));
	area += (p3->x) * ((p1->y) - (p2->y));
/*	area *= HALF; */
    RA(t, bpx_t_area) = area;

    return true;
  }

/*   ************************
5.0  Routinen fuer Tiefe Null
     ************************ */

static int
ResAndLNaught(PT *p)
{
	int		index;

	counter++;

	index = offset[0]+counter;
	lArray[index] = IP(p,bpx_p_l);
	indexPlusArray[index] = IP(p,bpx_p_counter);
	RA(p, BY) = uPlus[IP(p,bpx_p_counter)];
	
	return true;
}

static int
ValueOnNaught(PT *p)
{
	RA(p, BY) = bpxTauQ * RA(p, BY) + V(p, 0);

	return true;
}

/*   *******************
6.0  Routinen auf MKPlus
     ******************* */

static int
MarkPointMKPlus(PT *p)
{
	if ( (!(IP(p,bpx_p_mark))) && ((p->boundP)!=DIRICHLET) ) {
		nPlus++;
		IP(p,bpx_p_mark) = -nPlus;
	}

	return true;
}

static int
RestrictMKPlus(PT *p)
{
	int		index, indexPlus;

	if (IP(p,bpx_p_mark)>=0)
		return true;

	indexPlus = offsetPlus - IP(p,bpx_p_mark);
    IP(p,bpx_p_mark) = nMinus - IP(p,bpx_p_mark);
	index = offset[depth] + IP(p,bpx_p_mark);
	lArray[index] = IP(p,bpx_p_l);
	indexPlusArray[index] = IP(p,bpx_p_counter);
	IP(p,bpx_p_l) = depth - 1;
	IP(p,bpx_p_counter) = indexPlus;

	uPlus[indexPlus] += uPlus[indexPlusArray[index]];

	if(!areaKnown)
		sPlus[indexPlus] += sPlus[indexPlusArray[index]];

	return true;
}

static int
AddMKPlus(PT *p)
{
	if (IP(p,bpx_p_mark) >=0)
		return true;
	IP(p,bpx_p_mark) = nMinus - IP(p,bpx_p_mark);
	RA(p, BY) += V(p, depth);

	return true;
}

/*   ********************
7.0  Routinen auf MKMinus
     ******************** */

static int
MarkPointMKMinus(PT *p)
{
	if ( (!IP(p,bpx_p_mark)) && ((p->boundP)!=DIRICHLET) ){
		nMinus++;
		IP(p,bpx_p_mark) = nMinus;
	}

	return true;
}

static int 
RestrictMKMinus(PT *p,EDG *ed)
{
	PT		*p1 = ed->p1, *p2 = ed->p2;
	real		uVal, sVal;
	int		index;

	if ( (IP(p,bpx_p_mark)<=0) || (IP(p,bpx_p_mark) > nMinus) )
		return true;

	index = offset[depth]+IP(p,bpx_p_mark);
	lArray[index] = IP(p,bpx_p_l);
	indexPlusArray[index] = IP(p,bpx_p_counter);

	uVal = uPlus[IP(p,bpx_p_counter)];

	if (IP(p1,bpx_p_mark)) {
		index = offsetPlus + IP(p1,bpx_p_mark) - nMinus;
		uPlus[index] += HALF * uVal;
	}

	if (IP(p2,bpx_p_mark)) {
		index = offsetPlus + IP(p2,bpx_p_mark) - nMinus;
		uPlus[index] += HALF * uVal;
	}

	if (!areaKnown) {
		sVal = sPlus[IP(p,bpx_p_counter)];

		if (IP(p1,bpx_p_mark)) {
			index = offsetPlus + IP(p1,bpx_p_mark) - nMinus;
			sPlus[index] += HALF * sVal;
		}

		if (IP(p2,bpx_p_mark)) {
			index = offsetPlus + IP(p2,bpx_p_mark) - nMinus;
			sPlus[index] += HALF * sVal;
		}
	}

	return true;
}

static int
AddMKMinus(PT *p)
{
	if ( (IP(p,bpx_p_mark)<=0) || (IP(p,bpx_p_mark)>nMinus) )
		return true;

	RA(p, BY) += V(p, depth);

	return true;
}

/*   **************
8.0  Die Funktion V
     ************** */

static real
V(PT *p,int k)
{
	int		index;

	if (k == 0) {
		counter++;
		index = offset[0] + counter;
	} else
		index = offset[k] + IP(p,bpx_p_mark);

	return ValV(indexPlusArray[index], k, lArray[index]);
}

/*   --------------------------------------------------
8.1  Die eigentliche Berechnung von V, das Herz von BPX
     -------------------------------------------------- */

static real
ValV(int index, int k0, int k1)
{
	real		val;

	val = bpxTau1 / ( qData + bpxTau*(pow(4.0, (double) (k0))) );		
	if (k1 < maxDepth)
		val -= bpxTau1 / ( qData + bpxTau*(pow(4.0, (double) (k1+1))) );	
	
	val *= uPlus[index] / sPlus[index];
	
	return val;
}

/*   ******************
9.0  Die Kantenroutinen
     ****************** */

static int
MarkPointsOnEdge(EDG *ed)
{
	MarkPointMKPlus(ed->p1);
	MarkPointMKPlus(ed->p2);
	MarkPointMKMinus(PM(ed));

	return true;
}

static int
RestrictOnEdge(EDG *ed)
{
	RestrictMKPlus(ed->p1);
	RestrictMKPlus(ed->p2);
	RestrictMKMinus(PM(ed),ed);

	return true;
}
	
static int
AddOnEdges(EDG *ed)
{
	AddMKPlus(ed->p1);
	AddMKPlus(ed->p2);
	AddMKMinus(PM(ed));

	return true;
}

static int 
Interpolate(EDG *ed)
{
	RA(PM(ed), BY) = HALF * (RA(ed->p1, BY) + RA(ed->p2, BY));
	
	return true;
}

/*    *************************
10.0  Restriktion und Summation
      ************************* */

static void
Restrictions()
{
	InitArrays();
	ApplyP(InitPoint, notDirichlet);
	offset[maxDepth] = - 1;
	offsetPlus = actTriang->noOfPoints - 1;

	for (depth = maxDepth; depth > 0; depth--) {
		nPlus = 0;
		nMinus = 0;
		ApplyEDepth(MarkPointsOnEdge, depth - 1, allReds);
		ApplyEDepth(RestrictOnEdge, depth - 1, allReds);
		ApplyTDepth(UnmarkPoints, depth, all);

		offset[depth - 1] = offset[depth] + nMinus + nPlus;
		offsetPlus = offsetPlus + nPlus;
	}

	counter = 0;
	ApplyP(ResAndLNaught, initialNotDirichlet);

	return;
}

static void
SumUpBPXTerms()
{
	bpxTau1 = 1.0 + bpxTau;
	bpxTauQ = bpxTau / (qData + bpxTau);

	counter = 0;
	ApplyP(ValueOnNaught, initialNotDirichlet);
	for (depth = 1; depth <= maxDepth; depth++) {
		ApplyEDepth(Interpolate, depth - 1, allReds);
		nPlus = 0;
		nMinus = 0;
		ApplyEDepth(MarkPointsOnEdge, depth - 1, allReds);
		ApplyEDepth(AddOnEdges, depth - 1, allReds);
		ApplyTDepth(UnmarkPoints, depth, all);
	}

	return;
}

/*    ****************
11.0  Die Hauptroutine
      **************** */

				
static int
InitBpxMul()
  {
	if ((solveState->ansatz)!=linear)
	  ZIBStdOut("bpx only for linear ansatz\n");
	if (bpx_p_l==-1)
	  bpx_p_l = GetAccess(nodesState->accPoint,sizeof(int),sizeof(int),"l");
	if (bpx_p_counter==-1)
	  bpx_p_counter = GetAccess(nodesState->accPoint,sizeof(int),sizeof(int),"counter");
	if (bpx_p_mark==-1)
	  bpx_p_mark = GetAccess(nodesState->accPoint,sizeof(int),sizeof(int),"mark");
	if (bpx_t_area==-1)
	  bpx_t_area = GetAccess(nodesState->accTriangle,sizeof(real),sizeof(real),"area");

	if ((bpx_p_l==-1)||(bpx_p_counter==-1)||(bpx_p_mark==-1)||(bpx_t_area==-1))
	  {
	    ZIBStdOut("Can't initialize BPX, access denied\n");
		return false;
	  }

	ApplyP(InitFields, all);
	ApplyT(ComputeArea, allHist);
	areaKnown = false;

	return true;
  }

static int
bpxPcxMul(int x, int y)
{
	if ((bpx_p_l==-1)||(bpx_t_area==-1)||
		((solveState->ansatz)!=linear))
	  {
		NodeAssign(y,x);
		return true;
	  }
	BX = x;
	BY = y;

	NodeSetZero(BY);
	
	maxDepth = actTriang->maxDepth;
	Restrictions();
/*	Lev0DirectSol(BY,BY); */
	SumUpBPXTerms();
	areaKnown = true;

	return true;
}
